from pypom import (
    Page,
    Region,
)

from .meta import PyPOMFormMetaclass


BaseFormPage = PyPOMFormMetaclass(
    str('BaseFormPage'),
    (Page,),
    {
        '__doc__': 'This is the base page form class'
                   'for schema based page objects.'
    }
)


BaseFormRegion = PyPOMFormMetaclass(
    str('BaseFormRegion'),
    (Region,),
    {
        '__doc__': 'This is the base region form class'
                   'for schema based page objects.'
    }
)



import pytest


def test_meta_form_page(browser):
    """ test metaclass with pypom form page"""
    import colander

    from pypom_form.widgets import StringWidget

    class MyStringWidget(StringWidget):
        pass

    class BaseFormSchema(colander.MappingSchema):
        title = colander.SchemaNode(colander.String(),
                                    selector=('id', 'id1'))

    class SubFormSchema(BaseFormSchema):
        name = colander.SchemaNode(colander.String(),
                                   selector=('id', 'id2'),
                                   pypom_widget=MyStringWidget(
                                       options={'test': 1}))

    from pypom_form.form import BaseFormPage

    class SubFormPage(BaseFormPage):
        schema_factory = SubFormSchema

    subform = SubFormPage(browser)
    import mock

    with mock.patch(
            'pypom_form.widgets.StringWidget.get_input_element') \
            as get_input_element:
        get_input_element.configure_mock(**{'return_value.value': 'the title'})
        assert subform.title == 'the title'

    with mock.patch(
            'pypom_form.widgets.StringWidget.get_input_element') \
            as get_input_element:
        get_input_element.configure_mock(**{'return_value.value': 'the name'})
        assert subform.name == 'the name'


def test_meta_form_region(browser):
    """ test metaclass with pypom form region"""
    import colander

    from pypom_form.widgets import StringWidget

    class MyStringWidget(StringWidget):
        pass

    class BaseFormSchema(colander.MappingSchema):
        title = colander.SchemaNode(colander.String(),
                                    selector=('id', 'id1'))

    class SubFormSchema(BaseFormSchema):
        name = colander.SchemaNode(colander.String(),
                                   selector=('id', 'id2'),
                                   pypom_widget=MyStringWidget(
                                       options={'test': 1}))

    import pypom
    from pypom_form.form import BaseFormRegion

    class SubFormRegion(BaseFormRegion):
        schema_factory = SubFormSchema

    subform = SubFormRegion(pypom.Page(browser))
    import mock

    with mock.patch(
            'pypom_form.widgets.StringWidget.get_input_element') \
            as get_input_element:
        get_input_element.configure_mock(**{'return_value.value': 'the title'})
        assert subform.title == 'the title'

    with mock.patch(
            'pypom_form.widgets.StringWidget.get_input_element') \
            as get_input_element:
        get_input_element.configure_mock(**{'return_value.value': 'the name'})
        assert subform.name == 'the name'


def test_pypom_inner_attribute_form_page(browser):
    """ test pypom inner attribute """
    import colander

    from pypom_form.widgets import StringWidget

    class MyStringWidget(StringWidget):
        pass

    class BaseFormSchema(colander.MappingSchema):
        title = colander.SchemaNode(colander.String(),
                                    selector=('id', 'id1'))

    class SubFormSchema(BaseFormSchema):
        name = colander.SchemaNode(colander.String(),
                                   selector=('id', 'id2'),
                                   pypom_widget=MyStringWidget(
                                       options={'test': 1}))

    from pypom_form.form import BaseFormPage

    class SubFormPage(BaseFormPage):
        schema_factory = SubFormSchema

    subform = SubFormPage(browser)

    assert subform.__pypom__
    assert 'title' in subform.__pypom__
    assert 'name' in subform.__pypom__
    assert list(subform.__pypom__.keys()) == ['title', 'name']
    assert isinstance(subform.__pypom__['title'], colander.SchemaNode)
    assert isinstance(subform.__pypom__['name'], colander.SchemaNode)


def test_pypom_inner_attribute_form_region(browser):
    """ test pypom inner attribute """
    import colander

    from pypom_form.widgets import StringWidget

    class MyStringWidget(StringWidget):
        pass

    class BaseFormSchema(colander.MappingSchema):
        title = colander.SchemaNode(colander.String(),
                                    selector=('id', 'id1'))

    class SubFormSchema(BaseFormSchema):
        name = colander.SchemaNode(colander.String(),
                                   selector=('id', 'id2'),
                                   pypom_widget=MyStringWidget(
                                       options={'test': 1}))

    from pypom_form.form import BaseFormPage

    class SubFormPage(BaseFormPage):
        schema_factory = SubFormSchema

    subform = SubFormPage(browser)

    assert subform.__pypom__
    assert 'title' in subform.__pypom__
    assert 'name' in subform.__pypom__
    assert list(subform.__pypom__.keys()) == ['title', 'name']
    assert isinstance(subform.__pypom__['title'], colander.SchemaNode)
    assert isinstance(subform.__pypom__['name'], colander.SchemaNode)


def test_meta_form_page_widget_region(browser):
    """ test metaclass with pypom form page"""
    import colander

    from pypom_form.widgets import StringWidget

    class MyStringWidget(StringWidget):
        pass

    class BaseFormSchema(colander.MappingSchema):
        title = colander.SchemaNode(colander.String(),
                                    selector=('id', 'id1'))

    class SubFormSchema(BaseFormSchema):
        name = colander.SchemaNode(colander.String(),
                                   selector=('id', 'id2'),
                                   pypom_widget=MyStringWidget(
                                       options={'test': 1}))

    from pypom_form.form import BaseFormPage

    class SubFormPage(BaseFormPage):
        schema_factory = SubFormSchema

    subform = SubFormPage(browser)

    import mock

    with mock.patch(
            'pypom_form.widgets.BaseWidgetRegion.wait_for_region_to_load') \
            as wait_for_region_to_load:
        wait_for_region_to_load.configure_mock(**{'return_value': None})

        title_region = subform.getWidgetRegion('title')
        name_region = subform.getWidgetRegion('name')

        from pypom_form.widgets import BaseWidgetRegion
        assert isinstance(title_region, BaseWidgetRegion)
        assert isinstance(name_region, BaseWidgetRegion)


def test_meta_form_region_widget_region(browser):
    """ test metaclass with pypom form page"""
    import colander

    from pypom_form.widgets import StringWidget

    class MyStringWidget(StringWidget):
        pass

    class BaseFormSchema(colander.MappingSchema):
        title = colander.SchemaNode(colander.String(),
                                    selector=('id', 'id1'))

    class SubFormSchema(BaseFormSchema):
        name = colander.SchemaNode(colander.String(),
                                   selector=('id', 'id2'),
                                   pypom_widget=MyStringWidget(
                                       options={'test': 1}))

    import pypom
    from pypom_form.form import BaseFormRegion

    class SubFormRegion(BaseFormRegion):
        schema_factory = SubFormSchema

    subform = SubFormRegion(pypom.Page(browser))

    import mock

    with mock.patch(
            'pypom_form.widgets.BaseWidgetRegion.wait_for_region_to_load') \
            as wait_for_region_to_load:
        wait_for_region_to_load.configure_mock(**{'return_value': None})

        title_region = subform.getWidgetRegion('title')
        name_region = subform.getWidgetRegion('name')

        from pypom_form.widgets import BaseWidgetRegion
        assert isinstance(title_region, BaseWidgetRegion)
        assert isinstance(name_region, BaseWidgetRegion)


def test_mixed_page_region(browser):
    """ Test mixed/nested page and region """
    import colander

    from pypom_form.widgets import StringWidget

    class MyStringWidget(StringWidget):
        pass

    class BaseFormSchema(colander.MappingSchema):
        title = colander.SchemaNode(colander.String(),
                                    selector=('id', 'id1'))

    class SubFormSchema(BaseFormSchema):
        name = colander.SchemaNode(colander.String(),
                                   selector=('id', 'id2'),
                                   pypom_widget=MyStringWidget(
                                       options={'test': 1}))

    from pypom_form.form import BaseFormPage

    class SubFormPage(BaseFormPage):
        schema_factory = SubFormSchema

    from pypom_form.form import BaseFormRegion

    class SubFormRegion(BaseFormRegion):
        schema_factory = SubFormSchema

    region = SubFormRegion(SubFormPage(browser))

    import mock

    with mock.patch(
            'pypom_form.widgets.BaseWidgetRegion.wait_for_region_to_load') \
            as wait_for_region_to_load:
        wait_for_region_to_load.configure_mock(**{'return_value': None})

        assert region.getWidgetRegion('title') != \
            region.page.getWidgetRegion('title')
        assert region.getWidgetRegion('name') != \
            region.page.getWidgetRegion('name')


def test_mixed_page_region_page_reference(browser):
    """ Test mixed/nested page and region """
    import colander

    from pypom_form.widgets import StringWidget

    class MyStringWidget(StringWidget):
        pass

    class BaseFormSchema(colander.MappingSchema):
        title = colander.SchemaNode(colander.String(),
                                    selector=('id', 'id1'))

    class SubFormSchema(BaseFormSchema):
        name = colander.SchemaNode(colander.String(),
                                   selector=('id', 'id2'),
                                   pypom_widget=MyStringWidget(
                                       options={'test': 1}))

    from pypom_form.form import BaseFormPage

    class SubFormPage(BaseFormPage):
        schema_factory = SubFormSchema

    from pypom_form.form import BaseFormRegion

    class SubFormRegion(BaseFormRegion):
        schema_factory = SubFormSchema

    page = SubFormPage(browser)
    region = SubFormRegion(page)

    assert region.page == page

    import mock

    with mock.patch(
            'pypom_form.widgets.BaseWidgetRegion.wait_for_region_to_load') \
            as wait_for_region_to_load:
        wait_for_region_to_load.configure_mock(**{'return_value': None})

        assert region.getWidgetRegion('title').page == \
            region
        assert page.getWidgetRegion('title').page == \
            page


def test_meta_set(browser):
    import colander

    from pypom_form.widgets import StringWidget

    class MyStringWidget(StringWidget):
        pass

    class BaseFormSchema(colander.MappingSchema):
        title = colander.SchemaNode(colander.String(),
                                    selector=('id', 'id1'))

    class SubFormSchema(BaseFormSchema):
        name = colander.SchemaNode(colander.String(),
                                   selector=('id', 'id2'),
                                   pypom_widget=MyStringWidget(
                                       options={'test': 1}))

    import pypom
    from pypom_form.form import BaseFormRegion

    class SubFormRegion(BaseFormRegion):
        schema_factory = SubFormSchema

    subform = SubFormRegion(pypom.Page(browser))
    import mock

    setter_mock = mock.MagicMock(wraps=subform.__class__.title.fset)
    mock_property = subform.__class__.title.setter(setter_mock)
    with mock.patch.object(subform.__class__, 'title', mock_property):
        assert subform.set('title', 'the title') == subform
        setter_mock.assert_called_once_with(subform, 'the title')


def test_meta_update(browser):
    import colander

    from pypom_form.widgets import StringWidget

    class MyStringWidget(StringWidget):
        pass

    class BaseFormSchema(colander.MappingSchema):
        title = colander.SchemaNode(colander.String(),
                                    selector=('id', 'id1'))

    class SubFormSchema(BaseFormSchema):
        name = colander.SchemaNode(colander.String(),
                                   selector=('id', 'id2'),
                                   pypom_widget=MyStringWidget(
                                       options={'test': 1}))

    import pypom
    from pypom_form.form import BaseFormRegion

    class SubFormRegion(BaseFormRegion):
        schema_factory = SubFormSchema

    subform = SubFormRegion(pypom.Page(browser))
    import mock

    setter_mock = mock.MagicMock(wraps=subform.__class__.title.fset)
    mock_property = subform.__class__.title.setter(setter_mock)
    with mock.patch.object(subform.__class__, 'title', mock_property):
        assert subform.update(**{'title': 'the title'}) == subform
        setter_mock.assert_called_once_with(subform, 'the title')

        with pytest.raises(KeyError):
            subform.update(**{'another': 'another'})


def test_meta_raw_update(browser):
    import colander

    from pypom_form.widgets import StringWidget

    class MyStringWidget(StringWidget):
        pass

    class BaseFormSchema(colander.MappingSchema):
        title = colander.SchemaNode(colander.String(),
                                    missing='',
                                    selector=('id', 'id1'))

    class SubFormSchema(BaseFormSchema):
        age = colander.SchemaNode(colander.Int(),
                                  selector=('id', 'id2'),
                                  pypom_widget=MyStringWidget(
                                      options={'test': 1}))

    import pypom
    from pypom_form.form import BaseFormRegion

    class SubFormRegion(BaseFormRegion):
        schema_factory = SubFormSchema

    subform = SubFormRegion(pypom.Page(browser))
    import mock

    setter_mock = mock.MagicMock(wraps=subform.__class__.title.fset)
    mock_property = subform.__class__.title.setter(setter_mock)
    with mock.patch.object(subform.__class__, 'age', mock_property):
        assert subform.raw_update(**{'age': '12'}) == subform
        setter_mock.assert_called_once_with(subform, 12)

        with pytest.raises(KeyError):
            subform.update(**{'another': 'another'})



def test_serialize():
    from pypom_form.fields import ObjectType

    assert ObjectType.serialize(None, [1, 2, 3]) == [1, 2, 3]


def test_deserialize():
    from pypom_form.fields import ObjectType

    assert ObjectType.deserialize(None, [1, 2, 3]) == [1, 2, 3]



import pytest
import mock


def test_base_widget_1():
    """ Assert widget is None """
    from pypom_form.widgets import BaseWidget

    widget = BaseWidget()
    assert widget.field is None


def test_base_widget_2():
    """ Assert options are stored """
    from pypom_form.widgets import BaseWidget

    widget = BaseWidget(options={'test': 1})
    assert widget.options['test'] == 1


def test_base_widget_3():
    """ Assert getter method is required """
    from pypom_form.widgets import BaseWidget

    widget = BaseWidget()
    assert callable(widget.getter_factory)
    with pytest.raises(NotImplementedError):
        widget.getter_factory()


def test_base_widget_4():
    """ Assert setter method is required """
    from pypom_form.widgets import BaseWidget

    widget = BaseWidget()
    assert callable(widget.setter_factory)
    with pytest.raises(NotImplementedError):
        widget.setter_factory()


def test_string_widget_1():
    """ Assert getter is implemented """
    from pypom_form.widgets import StringWidget
    import colander

    widget_class = StringWidget

    mock_field = colander.SchemaNode(
        colander.String(),
        selector=('id', 'id1'),
    )

    widget = widget_class(field=mock_field)
    get_input_element_mock = mock.MagicMock()
    get_input_element_mock.configure_mock(**{
        'return_value.value': 'test ok'})
    widget.get_input_element = get_input_element_mock
    getter = widget.getter_factory()
    page_mock = mock.MagicMock()
    page_mock.__class__.getter = property(fget=getter)
    assert page_mock.getter == 'test ok'


def test_string_widget_2():
    """ Assert setter is implemented """
    from pypom_form.widgets import StringWidget
    import colander

    widget_class = StringWidget

    mock_field = colander.SchemaNode(
        colander.String(),
        selector=('id', 'id1'),
    )

    widget = widget_class(field=mock_field)
    get_input_element_mock = mock.MagicMock()
    widget.get_input_element = get_input_element_mock
    setter = widget.setter_factory()
    page_mock = mock.MagicMock()
    page_mock.__class__.setter = property(fset=setter)
    page_mock.setter = 'test ok'
    assert get_input_element_mock. \
        return_value.fill.assert_called_once_with('test ok') is None


def test_textarea_widget_1():
    """ Assert getter is implemented """
    from pypom_form.widgets import TextAreaWidget
    import colander

    widget_class = TextAreaWidget

    mock_field = colander.SchemaNode(
        colander.String(),
        selector=('id', 'id1'),
        pypom_widget=TextAreaWidget(),
    )

    widget = widget_class(field=mock_field)
    get_input_element_mock = mock.MagicMock()
    get_input_element_mock.configure_mock(**{
        'return_value.value': 'test ok'})
    widget.get_input_element = get_input_element_mock
    getter = widget.getter_factory()
    page_mock = mock.MagicMock()
    page_mock.__class__.getter = property(fget=getter)
    assert page_mock.getter == 'test ok'


def test_textarea_widget_2():
    """ Assert setter is implemented """
    from pypom_form.widgets import TextAreaWidget
    import colander

    widget_class = TextAreaWidget

    mock_field = colander.SchemaNode(
        colander.String(),
        selector=('id', 'id1'),
        pypom_widget=TextAreaWidget(),
    )

    widget = widget_class(field=mock_field)
    get_input_element_mock = mock.MagicMock()
    widget.get_input_element = get_input_element_mock
    setter = widget.setter_factory()
    page_mock = mock.MagicMock()
    page_mock.__class__.setter = property(fset=setter)
    page_mock.setter = 'test ok'
    assert get_input_element_mock. \
        return_value.fill.assert_called_once_with('test ok') is None


def test_checkbox_widget_1():
    """ Assert getter is implemented """
    from pypom_form.widgets import CheckboxWidget
    import colander

    widget_class = CheckboxWidget

    mock_field = colander.SchemaNode(
        colander.Bool(),
        selector=('id', 'id1'),
    )

    widget = widget_class(field=mock_field)
    get_input_element_mock = mock.MagicMock()
    get_input_element_mock.configure_mock(**{
        'return_value.checked': True})
    widget.get_input_element = get_input_element_mock
    getter = widget.getter_factory()
    page_mock = mock.MagicMock()
    page_mock.__class__.getter = property(fget=getter)

    assert page_mock.getter is True


def test_checkbox_widget_2():
    """ Assert setter is implemented. Set True, True """
    from pypom_form.widgets import CheckboxWidget
    import colander

    widget_class = CheckboxWidget

    mock_field = colander.SchemaNode(
        colander.Bool(),
        selector=('id', 'id1'),
    )

    widget = widget_class(field=mock_field)
    get_input_element_mock = mock.MagicMock()
    widget.get_input_element = get_input_element_mock
    setter = widget.setter_factory()
    page_mock = mock.MagicMock()
    page_mock.__class__.setter = property(fset=setter)

    page_mock.setter = True
    # let's set True twice
    page_mock.setter = True

    assert get_input_element_mock. \
        return_value.check.call_count == 2
    assert get_input_element_mock. \
        return_value.uncheck.call_count == 0


def test_checkbox_widget_3():
    """ Assert setter is implemented. Set True, False """
    from pypom_form.widgets import CheckboxWidget
    import colander

    widget_class = CheckboxWidget

    mock_field = colander.SchemaNode(
        colander.Bool(),
        selector=('id', 'id1'),
    )

    widget = widget_class(field=mock_field)
    get_input_element_mock = mock.MagicMock()
    widget.get_input_element = get_input_element_mock
    setter = widget.setter_factory()
    page_mock = mock.MagicMock()
    page_mock.__class__.setter = property(fset=setter)

    page_mock.setter = True
    # let's set False now
    page_mock.setter = False

    assert get_input_element_mock. \
        return_value.check.call_count == 1
    assert get_input_element_mock. \
        return_value.uncheck.call_count == 1


def test_integer_widget_1():
    """ Assert getter is implemented """
    from pypom_form.widgets import StringWidget
    import colander

    widget_class = StringWidget

    mock_field = colander.SchemaNode(
        colander.Int(),
        selector=('id', 'id1'),
    )
    widget = widget_class(field=mock_field)
    get_input_element_mock = mock.MagicMock()
    get_input_element_mock.configure_mock(**{
        'return_value.value': '1'})
    widget.get_input_element = get_input_element_mock
    getter = widget.getter_factory()
    page_mock = mock.MagicMock()
    page_mock.__class__.getter = property(fget=getter)
    assert page_mock.getter == 1


def test_integer_widget_2():
    """ Assert setter is implemented """
    from pypom_form.widgets import StringWidget
    import colander

    widget_class = StringWidget

    mock_field = colander.SchemaNode(
        colander.Int(),
        selector=('id', 'id1'),
    )
    get_input_element_mock = mock.MagicMock()
    widget = widget_class(field=mock_field)
    widget.get_input_element = get_input_element_mock
    setter = widget.setter_factory()
    page_mock = mock.MagicMock()
    page_mock.__class__.setter = property(fset=setter)
    page_mock.setter = 1
    assert get_input_element_mock. \
        return_value.fill.assert_called_once_with('1') is None


def test_float_widget_1():
    """ Assert getter is implemented """
    from pypom_form.widgets import StringWidget
    import colander

    widget_class = StringWidget

    mock_field = colander.SchemaNode(
        colander.Float(),
        selector=('id', 'id1'),
    )
    widget = widget_class(field=mock_field)
    get_input_element_mock = mock.MagicMock()
    get_input_element_mock.configure_mock(**{
        'return_value.value': '-1.1'})
    widget.get_input_element = get_input_element_mock
    getter = widget.getter_factory()
    page_mock = mock.MagicMock()
    page_mock.__class__.getter = property(fget=getter)
    assert page_mock.getter == -1.1


def test_float_widget_2():
    """ Assert setter is implemented """
    from pypom_form.widgets import StringWidget
    import colander

    widget_class = StringWidget

    mock_field = colander.SchemaNode(
        colander.Float(),
        selector=('id', 'id1'),
    )
    get_input_element_mock = mock.MagicMock()
    widget = widget_class(field=mock_field)
    widget.get_input_element = get_input_element_mock
    setter = widget.setter_factory()
    page_mock = mock.MagicMock()
    page_mock.__class__.setter = property(fset=setter)
    page_mock.setter = -1.1
    assert get_input_element_mock. \
        return_value.fill.assert_called_once_with('-1.1') is None


def test_region_widget_not_implemented(browser):
    """ Region widget """
    import colander

    from pypom_form.widgets import StringWidget

    class MyStringWidget(StringWidget):
        pass

    class BaseFormSchema(colander.MappingSchema):
        title = colander.SchemaNode(colander.String(),
                                    selector=('id', 'id1'))

    class SubFormSchema(BaseFormSchema):
        name = colander.SchemaNode(colander.String(),
                                   selector=('id', 'id2'),
                                   pwidget=MyStringWidget(
                                       options={'test': 1}))

    from pypom_form.form import BaseFormPage

    class SubFormPage(BaseFormPage):
        schema_factory = SubFormSchema

    subform = SubFormPage(browser)

    from pypom_form.widgets import BaseWidgetRegion

    with mock.patch(
            'pypom_form.widgets.BaseWidgetRegion.wait_for_region_to_load') \
            as wait_for_region_to_load:
        wait_for_region_to_load.configure_mock(**{'return_value': None})

        widget_region = BaseWidgetRegion(subform)

    with pytest.raises(NotImplementedError):
        widget_region.get_help()
    with pytest.raises(NotImplementedError):
        widget_region.get_validation_errors()


def test_region_widget_get_label(browser):
    """ Region widget """
    import colander

    from pypom_form.widgets import StringWidget

    class MyStringWidget(StringWidget):
        pass

    class BaseFormSchema(colander.MappingSchema):
        title = colander.SchemaNode(colander.String(),
                                    selector=('id', 'id1'))

    class SubFormSchema(BaseFormSchema):
        name = colander.SchemaNode(colander.String(),
                                   selector=('id', 'id2'),
                                   pwidget=MyStringWidget(
                                       options={'test': 1}))

    from pypom_form.form import BaseFormPage

    class SubFormPage(BaseFormPage):
        schema_factory = SubFormSchema

    subform = SubFormPage(browser)

    from pypom_form.widgets import BaseWidgetRegion

    with mock.patch(
            'pypom_form.widgets.BaseWidgetRegion.wait_for_region_to_load') \
            as wait_for_region_to_load:
        wait_for_region_to_load.configure_mock(**{'return_value': None})

        widget_region = BaseWidgetRegion(subform)

    widget_region.find_element = mock.MagicMock(**{
        'return_value.value': 'title'})
    assert widget_region.get_label() == 'title'


def test_widget_base_region_class():
    from pypom_form.widgets import BaseWidget
    from pypom_form.widgets import BaseWidgetRegion

    assert BaseWidget.region_class == BaseWidgetRegion

    widget = BaseWidget()

    assert widget.region_class == BaseWidgetRegion


def test_widget_base_region_class_custom():
    from pypom_form.widgets import BaseWidget
    import pypom

    widget = BaseWidget(region_class=pypom.Region)

    assert widget.region_class == pypom.Region


def test_widget_base_input_selector_raises():
    from pypom_form.widgets import BaseWidget

    widget = BaseWidget()

    with pytest.raises(NotImplementedError):
        widget.input_selector


def test_widget_string_input_selector():
    from pypom_form.widgets import StringWidget

    widget = StringWidget()

    assert widget.input_selector == ('tag', 'input')


def test_widget_checkbox_input_selector():
    from pypom_form.widgets import CheckboxWidget

    widget = CheckboxWidget()

    assert widget.input_selector == ('css', 'input[type="checkbox"]')


def test_widget_region_root_selector(browser):
    from pypom_form.widgets import BaseWidget
    from pypom_form.widgets import BaseWidgetRegion

    assert BaseWidget.region_class == BaseWidgetRegion

    field = mock.MagicMock(**{'selector': ('id', 'xyz')})
    widget = BaseWidget(field=field)
    import pypom
    page = pypom.Page(browser)

    with mock.patch(
            'pypom_form.widgets.BaseWidgetRegion.wait_for_region_to_load') \
            as wait_for_region_to_load:
        wait_for_region_to_load.configure_mock(**{'return_value': None})

        assert widget.getWidgetRegion(page)._root_locator == ('id', 'xyz')


def test_widget_region_widget_reference(browser):
    from pypom_form.widgets import BaseWidget
    from pypom_form.widgets import BaseWidgetRegion

    assert BaseWidget.region_class == BaseWidgetRegion

    field = mock.MagicMock(**{'selector': ('id', 'xyz')})
    widget = BaseWidget(field=field)
    import pypom
    page = pypom.Page(browser)

    with mock.patch(
            'pypom_form.widgets.BaseWidgetRegion.wait_for_region_to_load') \
            as wait_for_region_to_load:
        wait_for_region_to_load.configure_mock(**{'return_value': None})

        assert widget.getWidgetRegion(page).__pypom_widget__ == widget


def test_get_input_element(browser):
    from pypom_form.widgets import StringWidget

    widget = StringWidget()
    widget.getWidgetRegion = mock.MagicMock(
        **{'return_value.find_element.return_value': 'element'})
    import pypom
    page = pypom.Page(browser)
    widget.get_input_element(page) == 'element'


def test_get_input_element_no_container(browser):
    from pypom_form.widgets import StringWidget

    widget = StringWidget()
    widget.getWidgetRegion = mock.MagicMock(
        **{'return_value.find_element.return_value': None,
           'return_value.root': 'root'})
    import pypom
    page = pypom.Page(browser)
    widget.get_input_element(page) == 'root'


def test_widget_region_wait_timeout(browser):
    import colander

    from pypom_form.widgets import StringWidget

    class MyStringWidget(StringWidget):
        pass

    class BaseFormSchema(colander.MappingSchema):
        title = colander.SchemaNode(colander.String(),
                                    selector=('id', 'id1'))

    class SubFormSchema(BaseFormSchema):
        name = colander.SchemaNode(colander.String(),
                                   selector=('id', 'id2'),
                                   pwidget=MyStringWidget(
                                       options={'test': 1}))

    from pypom_form.form import BaseFormPage

    class SubFormPage(BaseFormPage):
        schema_factory = SubFormSchema

    subform = SubFormPage(browser)
    subform.timeout = 0

    from pypom_form.widgets import BaseWidgetRegion

    with mock.patch(
            'pypom_form.widgets.BaseWidgetRegion.root') \
            as root:
        root.configure_mock(**{'visible': False})

        from selenium.common.exceptions import TimeoutException
        with pytest.raises(TimeoutException):
            BaseWidgetRegion(subform)


def test_widget_region_wait_not_timeout(browser):
    import colander

    from pypom_form.widgets import StringWidget

    class MyStringWidget(StringWidget):
        pass

    class BaseFormSchema(colander.MappingSchema):
        title = colander.SchemaNode(colander.String(),
                                    selector=('id', 'id1'))

    class SubFormSchema(BaseFormSchema):
        name = colander.SchemaNode(colander.String(),
                                   selector=('id', 'id2'),
                                   pwidget=MyStringWidget(
                                       options={'test': 1}))

    from pypom_form.form import BaseFormPage

    class SubFormPage(BaseFormPage):
        schema_factory = SubFormSchema

    subform = SubFormPage(browser)
    subform.timeout = 0

    from pypom_form.widgets import BaseWidgetRegion

    with mock.patch(
            'pypom_form.widgets.BaseWidgetRegion.root') \
            as root:
        root.configure_mock(**{'visible': True})

        BaseWidgetRegion(subform)



from collections import OrderedDict

import colander

from .fields import is_readonly


def _widgets_mapping():
    """ circular import issues """
    from .widgets import (
        StringWidget,
        CheckboxWidget,
    )

    # typ -> default widget mapping
    WIDGETS_MAPPING = {
        colander.String: StringWidget,
        colander.Bool: CheckboxWidget,
        colander.Int: StringWidget,
        colander.Float: StringWidget,
    }
    return WIDGETS_MAPPING


def _getWidgetRegion(self, name):
    """ Return the widget region for the given name"""
    return self.__pypom__[name].pypom_widget.getWidgetRegion(self)


def _set(self, name, value):
    """ Set value for the given name with chained calls support """
    field = self.__pypom__[name]
    if is_readonly(field) is not True:
        setattr(self, name, value)
    return self


def _update(self, **values):
    """ Bulk page update with chained calls support.
        Updates fields considering the PyPOM fields order in order
        to support edit forms with fields that depends on other
        fields.
    """
    value_keys = values.keys()
    pypom_values = self.__pypom__.items()
    pypom_keys = map(lambda item: item[0], pypom_values)

    if not set(value_keys) <= set(pypom_keys):
        raise KeyError

    # values must be a subset of the available declared fields
    for key, node in pypom_values:
        # set values with the specific order matching with the
        # schema definition
        if key in value_keys and is_readonly(node) is not True:
            self.set(key, values[key])

    return self


def _raw_update(self, **raw_values):
    """ Bulk page update with chained calls support.
        Updates fields with raw values (not deserialized)
        considering the PyPOM fields order in order
        to support edit forms with fields that depends on other
        fields.
    """
    schema = self.schema_factory()
    values = schema.deserialize(raw_values)
    return self.update(**values)


class PyPOMFormMetaclass(type):
    """ This is the metaclass that empower the page or region
        form with dynamically generated getter and setter
        attributes depending on the declarative schema.

        Thanks to this metaclass you are able to set or access
        values driving your browser with ``page.title = 'title'``
        or ``page.title``.

        It add a ``getWidgetRegion`` method if you want to
        look up a region widget for advanced widget interactions
        accessing to region widget methods.
    """

    def __new__(cls, clsname, bases, dct):
        schema_factory = dct.get('schema_factory', None)

        if schema_factory:
            dct['__pypom__'] = OrderedDict()
            dct['getWidgetRegion'] = _getWidgetRegion
            dct['set'] = _set
            dct['update'] = _update
            dct['raw_update'] = _raw_update

            schema = schema_factory()
            WIDGETS_MAPPING = _widgets_mapping()

            for child in schema.children:
                widget = getattr(child, 'pypom_widget', None)
                if widget is None:
                    typ = child.typ
                    widget_factory = WIDGETS_MAPPING[typ.__class__]
                    child.pypom_widget = widget_factory(field=child)
                else:
                    # set the field reference
                    child.pypom_widget.field = child

                dct[child.name] = property(
                    fget=child.pypom_widget.getter_factory(),
                    fset=child.pypom_widget.setter_factory())

                dct['__pypom__'][child.name] = child

        return type.__new__(cls, clsname, bases, dct)



from .form import BaseFormRegion


class BaseWidgetRegion(BaseFormRegion):
    """
        This is the base widget region associated to a widget.

        You can lookup the base widget region from a page or region
        form with ``page.getWidgetRegion('field_name')`` in order
        to enable advanced widget interactions.

        For example you can get the field label, the help text,
        validation errors or if you have a complex widget you can
        expose the widget API. For example a multi selection widget
        with a filter feature or a reference field with a navigation
        popup with all the portal contents.

        This way you are not only able to set or get field values but
        you can incapsulate complex logics in the widget region and
        expose them on the page or region form ready to be used
    """
    label_selector = ('tag', 'label')

    def get_label(self):
        """ Return the widget label """
        element = self.find_element(*self.label_selector)
        return element and element.value or ''

    def get_help(self):
        """ Return the widget text help """
        raise NotImplementedError

    def get_validation_errors(self):
        """ Return the widget validation error """
        raise NotImplementedError

    def wait_for_region_to_load(self):
        """Wait for the page region to load."""
        if self.root is not None:
            self.wait.until(lambda s: self.root.visible)


class BaseWidget(object):
    """ This is the base widget. It is not intended to be used itself
        but you can use it as base class for your own widgets.

        You can associate widgets on schema level or the form metaclass
        will associate a fallback depending on the schema type if
        ``pypom_widget`` is missing.

        You can provide your own ``region_class`` on the schema definition
        of overriding the ``region_class`` class attribute in your own widget
        implementations.
    """

    region_class = BaseWidgetRegion

    def __init__(self, field=None, region_class=None, options={}):
        self.field = field
        if region_class is not None:
            self.region_class = region_class
        self.options = options

    @property
    def input_selector(self):
        """ Returns the input selector inside the field container.

            Most or times is a tuple with ``('tag', 'input')`` but
            it might change depending on the widget type.

            It's up to you providing the input selector that matches
            your input type.
        """
        raise NotImplementedError

    def get_input_element(self, page):
        """ Return the input element.

            If you provide a selector for the container of the input
            element, it will return the input element itself
            (preferred way if you want to use advanced widget
            features).

            Otherwise the region root will be returned
        """
        region = self.getWidgetRegion(page)
        element = region.find_element(*self.input_selector)
        element = element or element or region.root
        return element

    def getWidgetRegion(self, page):
        """ Returns a dynamic widget region containing the
            root selector.

            This is an internal method used by the page or
            region metaclass in order to be able to expose
            the widget region simply calling
            ``page.getWidgetRegion()``.

            It also sets a reference to the widget itself containing
            the widget options on the widget region.
        """
        region = self.region_class(page)
        region._root_locator = self.field.selector
        region.__pypom_widget__ = self
        return region

    def getter_factory(self):
        """ Returns a generated method to be attached on the
            PyPOM page or region.

            This is an internal method used by the page or
            region metaclass in order to be able to generate
            the getter method for the field.
        """
        raise NotImplementedError

    def setter_factory(self):
        """ Returns a generated method to be attached on the
            PyPOM page or region.

            This is an internal method used by the page or
            region metaclass in order to be able to generate
            the getter method for the field.
        """
        raise NotImplementedError

    def serialize(self, value):
        """ Serialize value.
            This method returns the value serialized (from model
            value to browser or widget internal representation).
        """
        return self.field.serialize(value)

    def deserialize(self, value):
        """ Deserialize value.
            This method returns the value serialized (from browser
            or internal widget representation to model).
        """
        return self.field.deserialize(value)


class StringWidget(BaseWidget):
    """ String widget """

    input_selector = ('tag', 'input')

    def getter_factory(self):
        def _getter(page):
            element = self.get_input_element(page)
            value = element.value
            return self.deserialize(value)
        return _getter

    def setter_factory(self):
        def _setter(page, value):
            value = self.serialize(value)
            element = self.get_input_element(page)
            element.fill(value)
        return _setter


class TextAreaWidget(StringWidget):
    """ TextArea widget """

    input_selector = ('tag', 'textarea')


class CheckboxWidget(BaseWidget):
    """ Checkbox widget """

    input_selector = ('css', 'input[type="checkbox"]')

    def getter_factory(self):
        def _getter(page):
            element = self.get_input_element(page)
            value = element.checked
            return self.deserialize(value)
        return _getter

    def setter_factory(self):
        def _setter(page, value):
            element = self.get_input_element(page)
            if value:
                element.check()
            else:
                element.uncheck()
        return _setter



import pytest


@pytest.fixture
def browser():
    import mock
    browser = mock.MagicMock()

    from zope.interface import alsoProvides
    from pypom.splinter_driver import ISplinter
    alsoProvides(browser, ISplinter)
    return browser



from setuptools import setup, find_packages

version = '0.2.0'


install_requires = [
    'PyPOM[splinter]>=1.1.1',
    'colander',
]

tests_require = [
    'tox',
    'pytest-cov',
    'mock',
]

docs_require = [
    'Sphinx',
    'sphinx_rtd_theme',
    ]

setup(name='pypom_form',
      version=version,
      description="pypom_form",
      long_description=open("README.rst").read() + "\n" +
      open("CHANGES.txt").read(),
      # Get more strings from
      # http://pypi.python.org/pypi?:action=list_classifiers
      classifiers=[
          "Programming Language :: Python",
          "Programming Language :: Python :: 2.7",
          "Programming Language :: Python :: 3.5",
          "Framework :: Pytest",
          "Topic :: Software Development :: Testing",
          "License :: OSI Approved :: Apache Software License",
          ],
      keywords='',
      author='Tierra QA Team',
      author_email='DLQA@tierratelematics.com',
      url='http://pypom-form.readthedocs.io/en/latest/',
      license='Apache License, Version 2.0',
      packages=find_packages(exclude=['ez_setup']),
      include_package_data=True,
      zip_safe=False,
      install_requires=install_requires,
      entry_points="""
      # -*- Entry points: -*-
      [console_scripts]
      """,
      extras_require={
          'docs': docs_require,
          'tests': tests_require,
          },
      )



"""
From https://github.com/Kotti/Kotti/blob/master/kotti/views/form.py
"""
import colander


def is_readonly(field):
    """
        Returns True if the given field is readonly, otherwise False.

        Field readonly (True)
        ---------------------

        >>> from mock import MagicMock
        >>> field = MagicMock()
        >>> field.readonly = True
        >>> assert field.readonly is True

        >>> is_readonly(field) is True
        True

        Field readonly (False)
        ----------------------

        >>> field.readonly = False
        >>> assert field.readonly is False

        >>> is_readonly(field) is False
        True

        Field readonly (default: False)
        -------------------------------

        >>> field = object()
        >>> from pytest import raises
        >>> with raises(AttributeError):
        ...     field.readonly

        >>> is_readonly(field) is False
        True
    """
    return getattr(field, 'readonly', False)


class ObjectType(colander.SchemaType):
    """ A type leaving the value untouched.
    """

    @staticmethod
    def serialize(node, value):
        return value

    @staticmethod
    def deserialize(node, value):
        return value
